from ewah.constants import EWAHConstants as EC
from ewah.hooks.rapidmail import EWAHRapidmailHook
from ewah.operators.base import EWAHBaseOperator


class EWAHGoogleAdsOperator(EWAHBaseOperator):
    _NAMES = ["rapidmail"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: False,
        EC.ES_SUBSEQUENT: False,
    }

    _CONN_TYPE = EWAHRapidmailHook.conn_type

    def __init__(self, endpoint=None, *args, **kwargs):
        self.endpoint = endpoint or kwargs.get("target_table_name")
        super().__init__(*args, **kwargs)

    def ewah_execute(self, context):
        for batch in self.source_hook.get_all_data(endpoint=self.endpoint):
            self.upload_data(batch)



import os
import yaml
import re
import datetime
import pytz

from jinja2 import Template
from io import StringIO

from airflow.utils.db import create_session
from airflow.models import Variable

try:
    from yaml import CLoader as base_loader, CDumper as Dumper
except ImportError:
    from yaml import Loader as base_loader, Dumper

"""
Extend the loader class to be able to reference files in .yaml files.
Adapted from https://stackoverflow.com/questions/528281/how-can-i-include-a-yaml-file-inside-another
Accessed 2020-07-28
"""


class Loader(base_loader):
    def __init__(self, stream):
        with create_session() as session:
            airflow_variables = {var.key: var.val for var in session.query(Variable)}
        self._root = os.path.split(stream.name)[0]
        ctx = {
            "env": os.environ,
            "airflow_variables": airflow_variables,
            "pytz": {name: getattr(pytz, name) for name in pytz.__all__},
            "datetime": {
                name: getattr(datetime, name)
                for name in ["date", "datetime", "time", "timedelta", "tzinfo"]
            },
            "re": {name: getattr(re, name) for name in re.__all__},
        }
        # Enable Jinja2 in the yaml files
        yaml_stream = StringIO(Template(stream.read()).render(ctx))
        yaml_stream.name = stream.name
        super().__init__(yaml_stream)

    def scalar_from_file(self, node):
        filename = os.path.join(self._root, self.construct_scalar(node))
        with open(filename, "r") as f:
            return f.read()

    def include_other_yml(self, node):
        filename = os.path.join(self._root, self.construct_scalar(node))
        with open(filename, "r") as f:
            return yaml.load(f, Loader)


Loader.add_constructor("!text_from_file", Loader.scalar_from_file)
Loader.add_constructor("!yml_from_file", Loader.include_other_yml)



from ewah.operators.base import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC

from ewah.hooks.hubspot import EWAHHubspotHook


class EWAHHubspotOperator(EWAHBaseOperator):
    _NAMES = ["hubspot"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: False,
    }

    _CONN_TYPE = EWAHHubspotHook.conn_type

    def __init__(
        self,
        *args,
        object=None,
        properties=None,
        exclude_properties=None,
        associations=None,
        **kwargs
    ):
        if object is None:
            object = kwargs.get("target_table_name")
        assert (
            object in EWAHHubspotHook.ACCEPTED_OBJECTS.keys()
        ), "Object {0} is invalid!".format(object)
        if object == "properties":
            kwargs["primary_key"] = ["name", "object_type"]
        else:
            kwargs["primary_key"] = "id"
        super().__init__(*args, **kwargs)
        self.object = object
        if isinstance(properties, str):
            properties = [properties]
        if isinstance(exclude_properties, str):
            exclude_properties = [exclude_properties]
        if isinstance(associations, str) and not associations == "all":
            associations = [associations]
        nonetype = type(None)
        assert isinstance(properties, (list, nonetype))
        assert isinstance(exclude_properties, (list, nonetype))
        assert isinstance(associations, (list, nonetype)) or associations == "all"
        self.properties = properties
        self.exclude_properties = exclude_properties
        self.associations = associations

    def ewah_execute(self, context):
        for batch in self.source_hook.get_data_in_batches(
            object=self.object,
            properties=self.properties,
            exclude_properties=self.exclude_properties,
            associations=self.associations,
        ):
            self.upload_data(batch)



from ewah.hooks.base import EWAHBaseHook

import requests


class EWAHAirflowHook(EWAHBaseHook):
    """Get Airflow Metadata from an Airflow installation via the stable API."""

    _ATTR_RELABEL = {
        "url": "host",
        "user": "login",
    }

    conn_name_attr = "ewah_airflow_conn_id"
    default_conn_name = "ewah_airflow_default"
    conn_type = "ewah_airflow"
    hook_name = "EWAH Airflow Connection"

    # Resolve some more complex endpoints
    _ENDPOINT_DAGRUNS = "dags/~/dagRuns"
    _ENDPOINTS = {
        "dagRuns": "dags/~/dagRuns",
        "taskInstance": "dags/~/dagRuns/~/taskInstances",
        "taskInstances": "dags/~/dagRuns/~/taskInstances",
    }

    _BASE_URL = "{0}/api/v1/{1}"

    @staticmethod
    def get_ui_field_behaviour():
        return {
            "hidden_fields": ["schema", "extra"],
            "relabeling": {
                "host": "URL (without endpoint, e.g. 'myairflowurl.com')",
                "login": "Basic Auth Username",
                "password": "Basic Auth Password",
                "port": "Port (only if using SSH)",
            },
        }

    @staticmethod
    def get_connection_form_widgets() -> dict:
        """Returns connection widgets to add to connection form"""
        from flask_appbuilder.fieldwidgets import BS3TextFieldWidget
        from wtforms import StringField

        return {
            "extra__ewah_airflow__ssh_conn_id": StringField(
                "SSH Connection ID to Airflow Server (optional)",
                widget=BS3TextFieldWidget(),
            ),
            "extra__ewah_airflow__protocol": StringField(
                # Note: must not use SelectField
                "Connection protocol (if using SSH or not specified in URL; one of: http, https)",
                widget=BS3TextFieldWidget(),
                default="http",
            ),
        }

    def get_data_in_batches(
        self, endpoint, page_size=100, batch_size=10000, data_from=None
    ):
        endpoint = self._ENDPOINTS.get(endpoint, endpoint)
        params = {}
        if data_from:
            assert endpoint == self._ENDPOINT_DAGRUNS
            # get DagRuns that ended since data_from
            params["end_date_gte"] = data_from.isoformat()
            params["order_by"] = "end_date"
        auth = requests.auth.HTTPBasicAuth(self.conn.login, self.conn.password)
        if self.conn.ssh_conn_id:
            ssh_hook = EWAHBaseHook.get_hook_from_conn_id(conn_id=self.conn.ssh_conn_id)
            ssh_host = self.conn.host or "localhost"
            ssh_host = ssh_host.replace("https://", "").replace("http://", "")
            ssh_port = self.conn.port
            if not ssh_port:
                if self.conn.protocol == "http":
                    ssh_port = 80
                else:
                    ssh_port = 443
            else:
                ssh_port = int(ssh_port)
            local_bind_address = ssh_hook.start_tunnel(ssh_host, ssh_port)
            host = "{2}://{0}:{1}".format(
                local_bind_address[0],
                str(local_bind_address[1]),
                self.conn.protocol or "http",
            )
        else:
            host = self.conn.host
            if not host.startswith("http"):
                host = self.conn.protocol + "://" + host
        url = self._BASE_URL.format(host, endpoint)
        params["limit"] = page_size
        params["offset"] = 0
        data = []
        i = 0
        while True:
            i += 1
            self.log.info("Making request {0} to {1}...".format(i, url))
            request = requests.get(url, params=params, auth=auth)
            assert request.status_code == 200, request.text
            response = request.json()
            keys = list(response.keys())
            if "total_entries" in keys:
                # Most endpoint use pagination + give "total_entries" for requests
                # The key to get the data from the response may differ from endpoint
                if keys[0] == "total_entries":
                    data_key = keys[1]
                else:
                    data_key = keys[0]
                if not response[data_key]:
                    # You know that you fetched all items if an empty list is returned
                    # (Note: total_entries is not reliable)
                    yield data
                    data = []
                    break
                data += response[data_key]
                if len(data) >= batch_size:
                    yield data
                    data = []
            else:
                # Rare endpoint that does not paginate (usually singletons)
                yield [response]
                break
            params["offset"] = params["offset"] + params["limit"]

        if self.conn.ssh_conn_id:
            ssh_hook.stop_tunnel()
            del ssh_hook



from ewah.constants import EWAHConstants as EC
from ewah.hooks.google_analytics import EWAHGoogleAnalyticsHook
from ewah.operators.base import EWAHBaseOperator

import json
from datetime import datetime, timedelta


class EWAHGAOperator(EWAHBaseOperator):
    _NAMES = ["ga", "google_analytics"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: False,
        EC.ES_INCREMENTAL: True,
        EC.ES_SUBSEQUENT: True,
    }

    _CONN_TYPE = EWAHGoogleAnalyticsHook.conn_type

    def __init__(
        self,
        view_id,
        dimensions,
        metrics,
        page_size=10000,
        include_empty_rows=True,
        sampling_level=None,
        *args,
        **kwargs
    ):
        if kwargs.get("primary_key"):
            raise Exception(
                "primary_key supplied, but the field is "
                + "auto-generated by the operator!"
            )

        # Set default to otherwise non-defaulted kwargs
        if not kwargs.get("load_data_from_relative"):
            kwargs["load_data_from_relative"] = timedelta(days=3)

        shorthand = "ga:"
        dimensions = [
            ("" if dim.startswith(shorthand) else shorthand) + dim for dim in dimensions
        ]
        metrics = [
            ("" if metric.startswith(shorthand) else shorthand) + metric
            for metric in metrics
        ]

        if kwargs.get("extract_strategy") == EC.ES_SUBSEQUENT:
            if "ga:dateHour" in dimensions:
                kwargs["subsequent_field"] = "dateHour"
            else:
                kwargs["subsequent_field"] = "date"

        kwargs.update({"primary_key": [dim[3:] for dim in dimensions]})

        self.view_id = view_id
        self.sampling_level = sampling_level
        self.dimensions = dimensions
        self.metrics = metrics
        self.page_size = page_size
        self.include_empty_rows = include_empty_rows

        super().__init__(*args, **kwargs)

        if not any(["ga:dateHour" in dimensions, "ga:date" in dimensions]):
            raise Exception("'date' or 'dateHour' must be a dimension!")

        if len(dimensions) > 7:
            raise Exception(
                (
                    "Can only fetch up to 7 dimensions!" + " Currently {0} Dimensions"
                ).format(
                    str(len(dimensions)),
                )
            )

        if len(metrics) > 10:
            raise Exception(
                (
                    "Can only fetch up to 10 metrics!" + " Currently {0} Dimensions"
                ).format(
                    str(len(metrics)),
                )
            )

        if self.page_size > 10000:
            raise Exception("Please specify a page size equal to or lower than 10000.")

    def ewah_execute(self, context):
        if (
            self.extract_strategy == EC.ES_SUBSEQUENT
            and self.test_if_target_table_exists()
        ):
            data_from = self.get_max_value_of_column(self.subsequent_field)
            if self.load_data_from_relative:
                data_from -= self.load_data_from_relative
        else:
            data_from = self.data_from

        if isinstance(data_from, datetime):
            # This is not the case on subsequent loads with 'date' as time dimension
            data_from = data_from.date()

        for batch in self.source_hook.get_data_in_batches(
            view_id=self.view_id,
            dimensions=self.dimensions,
            metrics=self.metrics,
            page_size=self.page_size,
            include_empty_rows=self.include_empty_rows,
            sampling_level=self.sampling_level,
            data_from=data_from,  # tbd: subsequent!
            data_until=(self.data_until or datetime.now()).date(),
            chunking_interval=self.load_data_chunking_timedelta
            or timedelta(days=7 * 13),
        ):
            self.upload_data(batch)



from airflow import DAG

from ewah.constants import EWAHConstants as EC
from ewah.utils.airflow_utils import datetime_utcnow_with_tz
from ewah.operators.base import EWAHBaseOperator
from ewah.uploaders import get_uploader

from collections.abc import Iterable
from copy import deepcopy
from croniter import croniter
from datetime import datetime, timedelta
from typing import Optional, Type, Callable, List, Tuple, Union

import re


def dag_factory_atomic(
    dag_name: str,
    dwh_engine: str,
    dwh_conn_id: str,
    start_date: datetime,
    el_operator: Type[EWAHBaseOperator],
    operator_config: dict,
    target_schema_name: str,
    target_schema_suffix: str = "_next",
    target_database_name: Optional[str] = None,
    default_args: Optional[dict] = None,
    schedule_interval: Optional[Union[str, timedelta]] = timedelta(days=1),
    end_date: Optional[datetime] = None,
    read_right_users: Optional[Union[List[str], str]] = None,
    additional_dag_args: Optional[dict] = None,
    additional_task_args: Optional[dict] = None,
    logging_func: Optional[Callable] = None,
    dagrun_timeout_factor: Optional[float] = None,
    task_timeout_factor: Optional[float] = None,
    **kwargs
) -> Tuple[DAG]:
    def raise_exception(msg: str) -> None:
        """Add information to error message before raising."""
        raise Exception("DAG: {0} - Error: {1}".format(dag_name, msg))

    logging_func = logging_func or print

    if kwargs:
        logging_func("unused config: {0}".format(str(kwargs)))

    additional_dag_args = additional_dag_args or {}
    additional_task_args = additional_task_args or {}

    if not read_right_users is None:
        if isinstance(read_right_users, str):
            read_right_users = [u.strip() for u in read_right_users.split(",")]
        if not isinstance(read_right_users, Iterable):
            raise_exception("read_right_users must be an iterable or string!")

    # Initialize parameter with default value here that can be updated depending
    # on schedule_interval conditions
    catchup = False

    # Allow using cron-style schedule intervals
    if isinstance(schedule_interval, str):
        assert croniter.is_valid(
            schedule_interval
        ), "schedule_interval is not valid a cron expression!"
    elif isinstance(schedule_interval, timedelta):
        catchup = True
        # fake catchup = True: between start_date and end_date is one schedule_interval
        # --> run the full refreshs every schedule_interval at the same time instead of
        # having a drift in execution time!
        if end_date:
            end_date = min(end_date, datetime_utcnow_with_tz())
        else:
            end_date = datetime_utcnow_with_tz()

        start_date += (
            int((end_date - start_date) / schedule_interval) - 1
        ) * schedule_interval

        # case 1: end_date = start_date + schedule_interval
        # if the division result is a precise integer, that implies a definite end_date
        # --> adjust to get exactly one schedule_interval delta between start_date and
        # end_date to have one last run available (that should have run before end_date)

        # case 2: end_date > (start_date + schedule_interval)
        # Airflow executes after data_interval_end - start_date has to be
        # between exactly 1 and below 2 time schedule_interval before end_date!
        # end_date - 2*schedule_interval < start_date <= end_date - schedule_interval

        # Make sure only one execution every runs scheduled but manual triggers work!
        end_date = start_date + 2 * schedule_interval - timedelta(seconds=1)

    if dagrun_timeout_factor:
        _msg = "dagrun_timeout_factor must be a number between 0 and 1!"
        assert isinstance(dagrun_timeout_factor, (int, float)) and (
            0 < dagrun_timeout_factor <= 1
        ), _msg
        # If schedule_interval is a string, we assume it is a cron and set a default
        if isinstance(schedule_interval, str):
            default_timeout = timedelta(hours=1)
            additional_dag_args["dagrun_timeout"] = additional_dag_args.get(
              "dagrun_timeout", default_timeout
            )
        else:
          additional_dag_args["dagrun_timeout"] = additional_dag_args.get(
              "dagrun_timeout", dagrun_timeout_factor * schedule_interval
          )

    if task_timeout_factor:
        additional_task_args["execution_timeout"] = additional_task_args.get(
            "execution_timeout", task_timeout_factor * schedule_interval
        )

    dag = DAG(
        dag_name,
        catchup=catchup,
        default_args=default_args,
        max_active_runs=1,
        schedule_interval=schedule_interval,
        start_date=start_date,
        end_date=end_date,
        **additional_dag_args,
    )

    kickoff, final = get_uploader(dwh_engine).get_schema_tasks(
        dag=dag,
        dwh_engine=dwh_engine,
        dwh_conn_id=dwh_conn_id,
        target_schema_name=target_schema_name,
        target_schema_suffix=target_schema_suffix,
        target_database_name=target_database_name,
        read_right_users=read_right_users,
        **additional_task_args,
    )

    base_config = deepcopy(additional_task_args)
    base_config.update(operator_config.get("general_config", {}))
    with dag:
        for table in operator_config["tables"].keys():
            table_config = deepcopy(base_config)
            table_config.update(operator_config["tables"][table] or {})
            table_config.update(
                {
                    "task_id": "extract_load_" + re.sub(r"[^a-zA-Z0-9_]", "", table),
                    "dwh_engine": dwh_engine,
                    "dwh_conn_id": dwh_conn_id,
                    "extract_strategy": table_config.get(  # Default to full refresh
                        "extract_strategy", None
                    )
                    or EC.ES_FULL_REFRESH,  # value can be given as None in table conf
                    "target_table_name": operator_config["tables"][table].get(
                        "target_table_name", table
                    ),
                    "target_schema_name": target_schema_name,
                    "target_schema_suffix": target_schema_suffix,
                    "target_database_name": target_database_name,
                }
            )
            # Atomic DAG only works with full refresh and subsequent strategies!
            assert table_config["extract_strategy"] in (
                EC.ES_FULL_REFRESH,
                EC.ES_SUBSEQUENT,
            )
            table_config["load_strategy"] = table_config.get(
                "load_strategy",
                EC.DEFAULT_LS_PER_ES[table_config["extract_strategy"]],
            )
            table_task = el_operator(**table_config)
            kickoff >> table_task >> final

    return (dag,)



from ewah.hooks.base import EWAHBaseHook

from airflow.utils.file import TemporaryDirectory

import os
from bson.objectid import ObjectId
from copy import deepcopy
from pymongo import MongoClient
from pymongo import ASCENDING as asc
from tempfile import NamedTemporaryFile


class EWAHMongoDBHook(EWAHBaseHook):
    _ATTR_RELABEL = {
        "default_database": "schema",
        "uri": "password",
        "username": "login",
    }

    conn_name_attr: str = "ewah_mongodb_conn_id"
    default_conn_name: str = "ewah_mongodb_default"
    conn_type: str = "ewah_mongodb"
    hook_name: str = "EWAH MongoDB Connection"

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        return {
            "hidden_fields": ["extra"],
            "relabeling": {
                "host": "Host",
                "schema": "Default database",
                "login": "Username",
                "password": "Uri OR password",
            },
        }

    @staticmethod
    def get_connection_form_widgets() -> dict:
        """Returns connection widgets to add to connection form"""
        from flask_appbuilder.fieldwidgets import (
            BS3TextFieldWidget,
            BS3PasswordFieldWidget,
        )
        from wtforms import StringField, PasswordField
        from ewah.utils.widgets import EWAHTextAreaWidget

        return {
            "extra__ewah_mongodb__conn_style": StringField(
                "Connection Style (one of: uri, credentials)",
                default="uri",
                widget=BS3TextFieldWidget(),
            ),
            "extra__ewah_mongodb__tls": StringField(
                "SSL / TLS?",
                default="no",
                widget=BS3TextFieldWidget(),
            ),
            "extra__ewah_mongodb__tls_insecure": StringField(
                "TLS: Allow insecure connections? Aka 'tlsInsecure'",
                default="no",
                widget=BS3TextFieldWidget(),
            ),
            "extra__ewah_mongodb__ssl_cert": StringField(
                "SSL Certificate",
                widget=EWAHTextAreaWidget(rows=12),
            ),
            "extra__ewah_mongodb__ssl_private": StringField(
                "SSL Private Key",
                widget=EWAHTextAreaWidget(rows=12),
            ),
            "extra__ewah_mongodb__ssl_password": StringField(
                "SSL Certificate / Private Key Password",
                widget=BS3PasswordFieldWidget(),
            ),
            "extra__ewah_mongodb__auth_source": StringField(
                "Auth Source", widget=BS3TextFieldWidget()
            ),
            "extra__ewah_mongodb__auth_mechanism": StringField(
                "Auth Mechanism", widget=BS3TextFieldWidget()
            ),
            "extra__ewah_mongodb__ssh_conn_id": StringField(
                "SSH Connection ID", widget=BS3TextFieldWidget()
            ),
        }

    @classmethod
    def get_cleaner_callables(cls):
        # overwrite me for cleaner callables that are always called
        def clean_data(row):
            for key, value in row.items():
                if isinstance(value, ObjectId):
                    row[key] = str(value)
            return row

        return clean_data

    @property
    def mongoclient(self):
        if not hasattr(self, "_mc"):
            if self.conn.ssh_conn_id:
                if self.conn.conn_style == "uri":
                    raise Exception("Cannot have SSH tunnel with uri connection type!")
                if not hasattr(self, "_ssh_hook"):
                    self._ssh_hook = EWAHBaseHook.get_hook_from_conn_id(
                        conn_id=self.conn.ssh_conn_id
                    )
                    self.local_bind_address = self._ssh_hook.start_tunnel(
                        self.conn.host, self.conn.port
                    )
            else:
                self.local_bind_address = (self.conn.host, self.conn.port)

            conn_kwargs = {"tz_aware": True}
            if self.conn.conn_style == "uri":
                conn_kwargs["host"] = self.conn.uri
            else:
                conn_kwargs["host"] = self.local_bind_address[0]
                conn_kwargs["port"] = self.local_bind_address[1]
                if self.conn.username:
                    conn_kwargs["username"] = self.conn.username
                if self.conn.password:
                    conn_kwargs["password"] = self.conn.password

            with TemporaryDirectory() as tmp_dir:
                if self.conn.tls.lower().startswith(("y", "t")):
                    conn_kwargs["tls"] = True
                with NamedTemporaryFile(dir=tmp_dir) as ssl_cert:
                    if self.conn.ssl_cert:
                        ssl_cert.write(self.conn.ssl_cert.encode())
                        if self.conn.ssl_private:
                            # Concatenate into the same file
                            ssl_cert.write(self.conn.ssl_private.encode())
                        ssl_cert.seek(0)
                        conn_kwargs["tlsCertificateKeyFile"] = os.path.abspath(
                            ssl_cert.name
                        )

                    if self.conn.ssl_password:
                        conn_kwargs[
                            "tlsCertificateKeyFilePassword"
                        ] = self.conn.ssl_password
                    if self.conn.tls_insecure.lower().startswith(("y", "t")):
                        conn_kwargs["tlsInsecure"] = True
                    if self.conn.auth_source:
                        conn_kwargs["authSource"] = self.conn.auth_source
                    if self.conn.auth_mechanism:
                        conn_kwargs["authMechanism"] = self.conn.auth_mechanism
                    self._mc = MongoClient(**conn_kwargs)

        return self._mc

    def get_data_in_batches(
        self,
        collection,
        database=None,
        filter_expression=None,
        batch_size: int = 100000,
    ):
        mongo_database = self.mongoclient.get_database(
            name=database or self.conn.default_database
        )
        mongo_collection = mongo_database[collection]
        self.log.info(
            "\n\nFetching data with filter expression:\n{0}\n\n".format(
                filter_expression
            )
        )
        filter_expression = deepcopy(filter_expression) or {}
        cursor = mongo_collection.find(filter_expression)

        data = []
        for document in cursor:
            data.append(document)
            if len(data) >= batch_size:
                yield data
                data = []
        if data:
            yield data

    def close(self):
        if hasattr(self, "_mc"):
            if hasattr(self, "_ssh_hook"):
                self._ssh_hook.stop_tunnel()
                del self._ssh_hook
            self._mc.close()
            del self._mc



from datetime import datetime, timedelta
from yfinance import download

from ewah.constants import EWAHConstants as EC
from ewah.operators.base import EWAHBaseOperator
from ewah.utils.airflow_utils import datetime_utcnow_with_tz


class EWAHFXYFinanceOperator(EWAHBaseOperator):
    _NAMES = ["fx_yfinance"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: True,
    }

    def __init__(
        self,
        currency_pair,  # iterable of length 2
        frequency="1d",  # 1d, 1wk, 1mo
        *args,
        **kwargs
    ):
        if not frequency in ("1d", "1wk", "1mo"):
            if frequency == "daily":
                frequency = "1d"
            elif frequency == "weekly":
                frequency = "1wk"
            elif frequency == "monthly":
                frequency = "1mo"
            else:
                raise Exception("Frequency must be one of: 1d, 1wk, 1mo")

        if not len(currency_pair) == 2:
            raise Exception(
                "currency_pair must be iterable of length 2 "
                + "containing the currency pair."
            )

        # No connection id required for this operator
        if kwargs.get("source_conn_id"):
            self.log.info(
                "source_conn_id is not required for operator! Ignoring argument."
            )
        kwargs["source_conn_id"] = None
        # If incremental, the primary key column is 'date'
        if kwargs.get("primary_key"):
            self.log.info("primary_key is fixed for this operator. Using the default.")
        kwargs["primary_key"] = ["date"]

        self.currency_pair = currency_pair
        self.frequency = frequency

        super().__init__(*args, **kwargs)

    def ewah_execute(self, context):
        data_from = self.data_from or context["dag"].start_date
        # yfinance requires one day extra to also show the current day
        data_until = (self.data_until or datetime_utcnow_with_tz()) + timedelta(days=1)

        format_str = "%Y-%m-%d"
        currency_str = "{0}{1}=X".format(*self.currency_pair)
        data = download(
            currency_str,
            data_from.strftime(format_str),
            data_until.strftime(format_str),
            self.frequency,
            auto_adjust=False
        )
        data_list = []
        for index, row in data.iterrows():
            data_dict = {
                # UNIX timestamp is calculated, adding 82800 to match unix timestamp from yahoofinancials (legacy)
                'date': int(datetime.strptime(index.strftime('%Y-%m-%d'), '%Y-%m-%d').timestamp() + 82800),
                'high': float(row['High'].iloc[0]),
                'low': float(row['Low'].iloc[0]),
                'open': float(row['Open'].iloc[0]),
                'close': float(row['Close'].iloc[0]),
                'volume': float(row['Volume'].iloc[0]),
                'adjclose': float(row['Adj Close'].iloc[0]),
                'formatted_date': index.strftime('%Y-%m-%d')
            }
            data_list.append(data_dict)

        self.upload_data(data_list)



from ewah.hooks.base import EWAHBaseHook


class EWAHdbtEnvVarHook(EWAHBaseHook):
    """
    Very simple hook to show a custom connection type in Airflow.
    Not used for data loading.
    Use it to create environment variable secrets using Airflow connections
    which are then used in dbt runs. Usage may be extended to other use cases
    of environment variables in the future.
    """

    _ATTR_RELABEL = {}

    conn_name_attr = "ewah_dbt_env_var_conn_id"
    default_conn_name = "ewah_dbt_env_var_default"
    conn_type = "ewah_dbt_env_var"
    hook_name = "EWAH dbt Environment Variable"

    @staticmethod
    def get_ui_field_behaviour():
        return {
            "hidden_fields": ["extra", "port", "host", "schema"],
            "relabeling": {
                "login": "Environment Variable Name",
                "password": "Environment Variable Value",
            },
        }



from ewah.operators.base import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC
from ewah.hooks.salesforce import EWAHSalesforceHook

from datetime import datetime
from typing import Optional


class EWAHSalesforceOperator(EWAHBaseOperator):
    _NAMES = ["sf", "salesforce"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: True,
        EC.ES_SUBSEQUENT: True,
    }

    _CONN_TYPE = EWAHSalesforceHook.conn_type

    _SUBSEQUENT_PLACEHOLDER = "~*subsequent-field-placeholder*~"

    def __init__(
        self, salesforce_object: Optional[str] = None, *args, **kwargs
    ) -> None:
        self.salesforce_object = salesforce_object or kwargs.get("target_table_name")
        kwargs["primary_key"] = "Id"
        # default subsequent_field to placeholder to be filled during execution
        if kwargs.get("extract_strategy") == EC.ES_SUBSEQUENT:
            kwargs["subsequent_field"] = kwargs.get(
                "subsequent_field", self._SUBSEQUENT_PLACEHOLDER
            )
        super().__init__(*args, **kwargs)

    def ewah_execute(self, context: dict) -> None:
        self.log.info(f"Fetching Salesforce data for {self.salesforce_object}...")
        is_subsequent = self.extract_strategy == EC.ES_SUBSEQUENT
        if is_subsequent:
            if self.subsequent_field == self._SUBSEQUENT_PLACEHOLDER:
                self.subsequent_field = self.source_hook.get_incrementer(
                    salesforce_object=self.salesforce_object
                )
            if self.test_if_target_table_exists():
                self.data_from = self.get_max_value_of_column(self.subsequent_field)

        for batch in self.source_hook.get_data_in_batches(
            salesforce_object=self.salesforce_object,
            data_from=self.data_from,
            data_until=self.data_until,
        ):
            if is_subsequent:
                for item in batch:
                    item[self.subsequent_field] = datetime.strptime(
                        item[self.subsequent_field], "%Y-%m-%dT%H:%M:%S.%f%z"
                    )
            self.upload_data(batch)
